#include	<stdio.h>
#include	<stdlib.h>
#include	<tsdb.h>
#include	<assert.h>
#include	<string.h>
#include	<getopt.h>
#include	<ace/mrs.h>
#include	<ace/tree.h>
#include	<unistd.h>

struct tsdb	*out;
int	parse_id_gen = 1;

char	*get(struct tsdb	*t, char	*rel, char	*field, char	*key, char	*id)
{
	struct relation	*r = get_relation(t, rel);
	char	**tup = tsdb_lookup_relation_with_key(r, get_field(r, key, "integer"), id);
	if(!tup)return NULL;
	return tup[get_field(r, field, "integer")];
}

char	**gettup2(struct relation	*r, char	*key1, char	*id1, char	*key2, char	*id2)
{
	char	**tup = NULL;
	int	f2 = get_field(r, key2, "integer");
	void filter(char	**t)
		{ if(!strcmp(t[f2], id2)) tup = t; }
	tsdb_visit_relation_with_key(r, get_field(r, key1, "integer"), id1, filter);
	return tup;
}

int	get_pref(struct tsdb	*t, char	*id, struct tree	**T, struct mrs	**M)
{
	char	*pid = get(t, "parse", "parse-id", "i-id", id);
	if(!pid)return -1;
	char	*tac = get(t, "tree", "t-active", "parse-id", pid);
	if(!tac)return -2;
	char	*rid = get(t, "preference", "result-id", "parse-id", pid);
	if(!rid)return -3;
	struct relation	*result = get_relation(t, "result");
	char	**tup = gettup2(result, "parse-id", pid, "result-id", rid);
	assert(tup);
	char	*deriv = tup[get_field(result, "derivation", "string")];
	char	*mrs = tup[get_field(result, "mrs", "string")];
	assert(deriv && mrs);
	*T = string_to_tree(strdup(deriv));
	if(!*T) { fprintf(stderr, "unable to parse derivation: %s\n", deriv); exit(-1); }
	*M = parse_mrs(strdup(mrs));
	if(!*M) { fprintf(stderr, "unable to parse MRS: %s\n", mrs); exit(-1); }
	assert(*T && *M);
	return 0;
}

void	add_tuple(struct relation	*r, char	**tup)
{
	int	j;

	for(j=0;j<r->nfields;j++)
	{
		if(!tup[j])
		{
			char	*type = r->fields[j].type;
			if(!strcasecmp(type, "integer"))tup[j] = strdup("-1");
			else if(!strcasecmp(type, "string"))tup[j] = strdup("");
			else if(!strcasecmp(type, "date"))tup[j] = strdup("23-6-2013 14:28:24");	// XXX fix me
			else
			{
				fprintf(stderr, "unknown tsdb type ':%s'\n", type);
				tup[j] = strdup("");
			}
		}
	}
	if(r->ntuples+1 > r->atuples)
	{
		r->atuples = (r->ntuples+1) * 1.5;
		r->tuples = realloc(r->tuples, sizeof(char**)*r->atuples);
	}
	r->tuples[r->ntuples] = malloc(sizeof(char*)*r->nfields);
	memcpy(r->tuples[r->ntuples], tup, sizeof(char*)*r->nfields);
	r->ntuples++;
}

purge_tuples(struct relation	*r, int	field, char	*value)
{
	int i, j;
	for(i=j=0;i<r->ntuples;i++)
	{
		if(!strcmp(r->tuples[i][field], value))
		{
			int k;
			for(k=0;k<r->nfields;k++)free(r->tuples[i][k]);
		}
		else r->tuples[j++] = r->tuples[i];
	}
	r->ntuples = j;
}

int	reindex_and_write(struct tsdb	*t, struct relation	*r)
{
	tsdb_reindex_relation(r);
	int res = tsdb_write_relation(t, r);
	if(res != 0)
	{
		fprintf(stderr, "failed to write '%s' relation.\n", r->name);
		return -1;
	}
	return 0;
}

int	len_of_derivation(struct tree	*t)
{
	int	len = 2*strlen(t->label) + 100;
	int i;
	for(i=0;i<t->ndaughters;i++)
		len += len_of_derivation(t->daughters[i]);
	for(i=0;i<t->ntokens;i++)
	{
		len += 10;
		len += strlen(t->tokens[i]);
	}
	return len;
}

char	*dqescape(char *s)
{
	char *s2 = malloc(strlen(s)*2+1), *p = s2;
	while(*s)
	{
		if(*s=='\b') { *p++ = '\\'; *p++ = 'b'; s++; continue; }
		if(*s=='\f') { *p++ = '\\'; *p++ = 'f'; s++; continue; }
		if(*s=='\n') { *p++ = '\\'; *p++ = 'n'; s++; continue; }
		if(*s=='\r') { *p++ = '\\'; *p++ = 'r'; s++; continue; }
		if(*s=='\t') { *p++ = '\\'; *p++ = 't'; s++; continue; }
		// specs for json apparently say you should escape '/'...
		if(*s == '\'' || *s == '\\' || *s == '"' || *s=='/')*p++='\\';
		*p++ = *s++;
	}
	*p++ = 0;
	return s2;
}

int	edge_id_gen = 0;

char	*build_derivation(char	*str, struct tree	*t)
{
	int i;
	if(t->ndaughters)
	{
		t->edge_id = edge_id_gen++;
		str += sprintf(str, "(%d %s %f %d %d", t->edge_id, t->label, 0.0, t->tfrom, t->tto);
		for(i=0;i<t->ndaughters;i++)
		{
			*str++ = ' ';
			str = build_derivation(str, t->daughters[i]);
		}
	}
	else
	{
		char	*el = dqescape(t->label);
		str += sprintf(str, "(\"%s\"", el);
		free(el);
		for(i=0;i<t->ntokens;i++)
		{
			char	*e = dqescape(t->tokens[i]);
			str += sprintf(str, " %d \"%s\"", t->tokenids[i], e);
			free(e);
		}
	}
	*str++ = ')';
	*str = 0;
	return str;
}

char	*tree_to_derivation(struct tree	*t)
{
	int	len = len_of_derivation(t);
	char	*str = malloc(len+1);
	build_derivation(str, t);
	return str;
}

char	*mrs_to_string(struct mrs	*m)
{
	extern char	*mrs_end;
	char	*save_end;
	save_end = mrs_end;
	mrs_end = "";
	int	l = snprint_mrs(NULL, 0, m);
	char	*str = malloc(l+10);
	snprint_mrs(str, l+1, m);
	mrs_end = save_end;
	return str;
}

record(char	*id, int	reject, int	n, struct tree	**t, struct mrs	**m)
{
	char	pid[128];
	char	readings[128];
	int	hallucinate_rejection_tree = (n && reject)?1:0;
	sprintf(pid, "%d", parse_id_gen++);
	sprintf(readings, "%d", n + hallucinate_rejection_tree);

	struct relation	*parse = get_relation(out, "parse");
	purge_tuples(parse, get_field(parse, "i-id", "integer"), id);
	char	*ptup[parse->nfields];
	bzero(ptup, sizeof(ptup));
	ptup[get_field(parse, "i-id", "integer")] = strdup(id);
	ptup[get_field(parse, "parse-id", "integer")] = strdup(pid);
	ptup[get_field(parse, "readings", "integer")] = strdup(readings);
	add_tuple(parse, ptup);

	struct relation	*result = get_relation(out, "result");
	purge_tuples(result, get_field(result, "parse-id", "integer"), pid);
	int i = 0;
	edge_id_gen = 1;
	for(i=0;i<n;i++)
	{
		char	*rtup[result->nfields];
		bzero(rtup, sizeof(rtup));
		rtup[get_field(result, "parse-id", "integer")] = strdup(pid);
		char	rid[128];
		sprintf(rid, "%d", i);
		rtup[get_field(result, "result-id", "integer")] = strdup(rid);
		char	*deriv = tree_to_derivation(t[i]);
		rtup[get_field(result, "derivation", "string")] = strdup(deriv);
		char	*mrs = mrs_to_string(m[i]);
		rtup[get_field(result, "mrs", "string")] = strdup(mrs);
		add_tuple(result, rtup);
	}
	if(hallucinate_rejection_tree)
	{
		char	*rtup[result->nfields];
		bzero(rtup, sizeof(rtup));
		rtup[get_field(result, "parse-id", "integer")] = strdup(pid);
		char	rid[128];
		sprintf(rid, "%d", i);
		rtup[get_field(result, "result-id", "integer")] = strdup(rid);
		rtup[get_field(result, "derivation", "string")] = strdup("(0 rejected-all-analyses 0.00 0 1 (\"reject\" 0 1))");
		rtup[get_field(result, "tree", "string")] = strdup("(\"reject\" (\"reject\"))");
		rtup[get_field(result, "mrs", "string")] = strdup("[ LTOP: h1 INDEX: e2 RELS: < > HCONS: < > ]");
		add_tuple(result, rtup);
	}
	if(!reject && n==1)
	{
		// harmoniously accept -- write a 'tree' record and a 'preference' record
		struct relation	*pref = get_relation(out, "preference");
		struct relation	*tree = get_relation(out, "tree");

		char	*tup[pref->nfields];
		bzero(tup, sizeof(tup));
		tup[get_field(pref, "parse-id", "integer")] = strdup(pid);
		tup[get_field(pref, "result-id", "integer")] = strdup("0");
		add_tuple(pref, tup);

		char	*ttup[tree->nfields];
		bzero(ttup, sizeof(ttup));
		ttup[get_field(tree, "parse-id", "integer")] = strdup(pid);
		ttup[get_field(tree, "t-active", "integer")] = strdup("1");
		add_tuple(tree, ttup);
	}
}

int	same_tree(struct tree	*x, struct tree	*y)
{
	if(strcmp(x->label, y->label))return 0;
	if(x->ndaughters==1 && y->ndaughters==1
		&& x->daughters[0]->ndaughters == 0
		&& y->daughters[0]->ndaughters == 0)
		return 1;
	if(x->ndaughters != y->ndaughters)return 0;
	int i;
	for(i=0;i<x->ndaughters;i++)
		if(!same_tree(x->daughters[i], y->daughters[i]))
			return 0;
	return 1;
}

int	same_mrs(struct mrs	*x, struct mrs	*y)
{
	if(mrs_subsumes(x,y) && mrs_subsumes(y,x))return 1;
	else return 0;
}

main(int	argc, char	**argv)
{
	char	*grampath = NULL;
	char	*outpath = NULL;
	char	*idpath = NULL;
	int i, ch;
	while( (ch = getopt(argc, argv, "g:o:i:")) != -1) switch(ch)
	{
		case	'g':	grampath = optarg; break;
		case	'o':	outpath = optarg; break;
		case	'i':	idpath = optarg; break;
	}
	assert(grampath);
	assert(outpath);
	assert(idpath);

	ace_load_grammar(grampath);

	assert(argc > optind);

	char	cmd[1024];
	sprintf(cmd, "./mkprof -s %s %s\n", argv[optind], outpath);
	system(cmd);
	out = load_tsdb(outpath);
	assert(out);

	int			nfrom = argc-optind;
	struct tsdb	**from = calloc(sizeof(struct tsdb*),nfrom);
	char		**nickname = calloc(sizeof(char*),nfrom);
	for(i=0;i<nfrom;i++)
	{
		from[i] = load_tsdb(argv[optind+i]);
		nickname[i] = strrchr(argv[optind+i],'/');
		if(nickname[i] && nickname[i][1])nickname[i]++;
		else asprintf(&nickname[i], "src-%d", i);
		assert(from[i]);
	}

	FILE	*f=fopen(idpath,"r");
	assert(f);
	char	id[128];
	int	n_total = 0;
	int	n_unanimously_reject = 0;
	int	n_discordantly_reject = 0;
	int	n_discordantly_accept = 0;
	int	n_harmoniously_accept = 0;
	while(fscanf(f, "%s", id) == 1 && !feof(f))
	{
		printf("id = %s\n", id);
		clear_slab();
		clear_mrs();
		int j;
		int	no = 0, nr = 0;
		struct tree	**to = NULL;
		struct mrs	**mo = NULL;
		int		ido[nfrom];
		for(j=0;j<nfrom;j++)
		{
			struct tree	*t;
			struct mrs	*m;
			int result = get_pref(from[j], id, &t, &m);
			if(result == 0)
			{
				int k;
				for(k=0;k<no;k++)
				{
					if(same_mrs(m, mo[k]))
					{
						printf("%12s: = %s\n", nickname[j], nickname[ido[k]]);
						break;
					}
				}
				if(k==no)
				{
					printf("%12s: new\n", nickname[j]);
					no++;
					to = realloc(to, sizeof(struct tree*)*no);
					mo = realloc(mo, sizeof(struct mrs*)*no);
					to[no-1] = t;
					mo[no-1] = m;
					ido[no-1] = j;
				}
			}
			else
			{
				printf("%12s: reject\n", nickname[j]);
				nr++;
			}
		}
		printf("%d rejections, %d analyses\n", nr, no);

		n_total++;
		if(no==0)n_unanimously_reject++;
		else if(nr)n_discordantly_reject++;
		else if(no>1)n_discordantly_accept++;
		else n_harmoniously_accept++;

		record(id, nr, no, to, mo);
	}
	printf("---------\n");
	printf("total items:         %d\n", n_total);
	printf("unanimously reject:  %d = %.1f%%\n", n_unanimously_reject, 100*(double)n_unanimously_reject / n_total);
	printf("discordantly reject: %d = %.1f%%\n", n_discordantly_reject, 100*(double)n_discordantly_reject / n_total);
	printf("discordantly accept: %d = %.1f%%\n", n_discordantly_accept, 100*(double)n_discordantly_accept / n_total);
	printf("harmoniously accept: %d = %.1f%%\n", n_harmoniously_accept, 100*(double)n_harmoniously_accept / n_total);

	reindex_and_write(out, get_relation(out, "parse"));
	reindex_and_write(out, get_relation(out, "result"));
	reindex_and_write(out, get_relation(out, "tree"));
	reindex_and_write(out, get_relation(out, "preference"));
}
