#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<assert.h>

#include	<tsdb.h>
#include	<ace/tree.h>

int debug_compare = 0;

int	same_tree(struct tree	*x, struct tree	*y)
{
	if(strcasecmp(x->label, y->label))
	{
		if(debug_compare)
			printf("%s is not %s\n", x->label, y->label);
		return 0;
	}
	if(x->ndaughters != y->ndaughters)
	{
		if(debug_compare)
			printf("wrong arity (%d vs %d)\n", x->ndaughters, y->ndaughters);
		return 0;
	}
	int i;
	for(i=0;i<x->ndaughters;i++)
		if(!same_tree(x->daughters[i], y->daughters[i]))
			return 0;
	return 1;
}

int	same_derivation(char	*d1, char	*d2)
{
	if(debug_compare)printf("compare derivations\n");
	d1 = strdup(d1);
	d2 = strdup(d2);
	struct tree	*t1 = string_to_tree(d1);
	struct tree	*t2 = string_to_tree(d2);
	if(debug_compare)
	{
		printf("LHS:\n");print_tree(t1,0);printf("\n");
		printf("RHS:\n");print_tree(t2,0);printf("\n");
	}
	free(d1);
	free(d2);
	int result = same_tree(t1, t2);
	free_tree(t1);
	free_tree(t2);
	return result;
}

void	add_tuple(struct relation	*r, char	**tup)
{
	int	j;

	for(j=0;j<r->nfields;j++)if(!tup[j])tup[j] = strdup("");
	if(r->ntuples+1 > r->atuples)
	{
		r->atuples = (r->ntuples+1) * 1.5;
		r->tuples = realloc(r->tuples, sizeof(char**)*r->atuples);
	}
	r->tuples[r->ntuples] = malloc(sizeof(char*)*r->nfields);
	memcpy(r->tuples[r->ntuples], tup, sizeof(char*)*r->nfields);
	r->ntuples++;
}

int	reindex_and_write(struct tsdb	*t, struct relation	*r)
{
	tsdb_reindex_relation(r);
	int res = tsdb_write_relation(t, r);
	if(res != 0)
	{
		fprintf(stderr, "failed to write '%s' relation.\n", r->name);
		return -1;
	}
	return 0;
}

store_annotation(struct tsdb	*parsed, char	*pid, char	*rid)
{
	struct relation	*tree = get_relation(parsed, "tree");

	int	tree_parse_id = get_field(tree, "parse-id", "integer");
	int tree_t_active = get_field(tree, "t-active", "integer");
	char	*treetup[tree->nfields];
	bzero(treetup, sizeof(treetup));
	treetup[tree_parse_id] = strdup(pid);
	treetup[tree_t_active] = strdup("1");
	add_tuple(tree, treetup);
	reindex_and_write(parsed, tree);

	if(rid)
	{
		struct relation	*pref = get_relation(parsed, "preference");

		int	pref_parse_id = get_field(pref, "parse-id", "integer");
		int pref_result_id = get_field(pref, "result-id", "integer");
		char	*preftup[pref->nfields];
		bzero(preftup, sizeof(preftup));
		preftup[pref_parse_id] = strdup(pid);
		preftup[pref_result_id] = strdup(rid);
		add_tuple(pref, preftup);
		reindex_and_write(parsed, pref);
	}
}

main(int	argc, char	*argv[])
{
	if(argc != 3)
	{
		fprintf(stderr, "usage: %s reparsed-profile gold-profile\n", argv[0]);
		fprintf(stderr, "populates the `tree' and `preference' relations in `reparsed-profile' by matching the preferred derivation annotated in `gold-profile', when present\n");
		exit(-1);
	}
	struct tsdb	*parsed, *gold;
	parsed = load_tsdb(argv[1]);
	gold = load_tsdb(argv[2]);
	assert(parsed && gold);

	// go through gold profile and find preferred items;
	// put a preferred annotation on corresponding parsed items
	struct relation	*gitem = get_relation(gold, "item");
	int	g_item_i_id = get_field(gitem, "i-id", "integer");
	int	g_item_i_input = get_field(gitem, "i-input", "string");
	struct relation	*gparse = get_relation(gold, "parse");
	int	g_parse_i_id = get_field(gparse, "i-id", "integer");
	int	g_parse_parse_id = get_field(gparse, "parse-id", "integer");
	struct relation	*gpref = get_relation(gold, "preference");
	int	g_pref_parse_id = get_field(gpref, "parse-id", "integer");
	int	g_pref_result_id = get_field(gpref, "result-id", "integer");
	struct relation	*gresult = get_relation(gold, "result");
	int	g_result_parse_id = get_field(gresult, "parse-id", "integer");
	int	g_result_result_id = get_field(gresult, "result-id", "integer");
	int	g_result_derivation = get_field(gresult, "derivation", "string");

	struct relation	*pitem = get_relation(parsed, "item");
	int	p_item_i_id = get_field(pitem, "i-id", "integer");
	int	p_item_i_input = get_field(pitem, "i-input", "string");
	struct relation	*pparse = get_relation(parsed, "parse");
	int	p_parse_i_id = get_field(pparse, "i-id", "integer");
	int	p_parse_parse_id = get_field(pparse, "parse-id", "integer");
	struct relation	*ppref = get_relation(parsed, "preference");
	int	p_pref_parse_id = get_field(ppref, "parse-id", "integer");
	int	p_pref_result_id = get_field(ppref, "result-id", "integer");
	struct relation	*presult = get_relation(parsed, "result");
	int	p_result_parse_id = get_field(presult, "parse-id", "integer");
	int	p_result_result_id = get_field(presult, "result-id", "integer");
	int	p_result_derivation = get_field(presult, "derivation", "string");

	int i;

	int	ngold = 0, nmatch = 0, nmiss = 0;
	for(i=0;i<gpref->ntuples;i++)
	{
		char	*gpid = gpref->tuples[i][g_pref_parse_id];
		char	*grid = gpref->tuples[i][g_pref_result_id];
		char	**gparse_tuple = tsdb_lookup_relation_with_key(gparse, g_parse_parse_id, gpid);
		assert(gparse_tuple != NULL);
		char	*giid = gparse_tuple[g_parse_i_id];
		//if(strcmp(giid, "3063951"))continue;
		printf("gold: item %s parse %s prefers result %s\n", giid, gpid, grid);
		char	*gderiv = NULL;
		void	visitor(char	**gresult_tuple)
		{
			if(!strcmp(gresult_tuple[g_result_result_id], grid))
				gderiv = gresult_tuple[g_result_derivation];
		}
		tsdb_visit_relation_with_key(gresult, g_result_parse_id, gpid, visitor);
		assert(gderiv != NULL);
		printf("   gold derivation has length %d\n", (int)strlen(gderiv));

		char	**pparse_tuple = tsdb_lookup_relation_with_key(pparse, p_parse_i_id, giid);
		assert(pparse_tuple != NULL);
		char	*ppid = pparse_tuple[p_parse_parse_id];
		printf("   in parsed profile, pid = %s\n", ppid);
		char	*prid = NULL;
		void	visitor2(char	**presult_tuple)
		{
			if(debug_compare)printf("checking rid %s\n", presult_tuple[p_result_result_id]);
			if(same_derivation(presult_tuple[p_result_derivation], gderiv))
				prid = presult_tuple[p_result_result_id];
		}
		tsdb_visit_relation_with_key(presult, p_result_parse_id, ppid, visitor2);
		ngold++;
		if(prid)
		{
			printf("  in parsed profile, rid = %s\n", prid);
			nmatch++;
		}
		else
		{
			printf("  in parsed profile, missing\n");
			nmiss++;
		}
		store_annotation(parsed, ppid, prid);
	}
	fprintf(stderr, "%d gold trees ; %d matches and %d misses\n", ngold, nmatch, nmiss);
}
