#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>
#include	<string.h>
#include	<unistd.h>
#include	<locale.h>
#include	<sys/fcntl.h>
#include	<sys/termios.h>

#include	<getopt.h>

#include	<liba.h>
#include	<ace/tdl.h>
#include	<tsdb.h>
#include	"hash.h"
#include	"version.h"

#define	ARBITER_PORT	9976

#define	TOKEN_EDGE_TYPE		0
#define	LEXEME_EDGE_TYPE	1
#define	LEXRULE_EDGE_TYPE	2
#define	SYNRULE_EDGE_TYPE	3
#define	ROOT_EDGE_TYPE		4

#define	RESULT_EDGE_STATUS	1
#define	ACTIVE_EDGE_STATUS	2
#define	PACKED_EDGE_STATUS	4
#define	FROSTED_EDGE_STATUS	8
#define	FROZEN_EDGE_STATUS	16

void	record_parse(char	*parse_id, char	*run_id, char	*i_id, int	nresults, int	*nedges);
void	record_run(char	*run_id, char	*platform, int	nitems);
void	record_result(char	*parse_id, int	result_idx, char	*derivation, char	*mrs, char	*surface, int	got_prob, double	prob);
void	record_edge(char	*parse_id, int	id, char	*sign, int	from, int	to, int	type, int	status, char	*dlist, char	*plist);

char	*get_source_mrs_for_i_id(char	*i_id);

char	*arbiter_host;
int		arbiter_fd;
FILE	*arbiter_f;
FILE	*arbiter_wrf;

char	*source_profile_path;
struct tsdb	*source_profile;
struct relation	*s_parse;
struct relation	*s_preference;
struct relation	*s_result;

char	*profile_path;
struct tsdb	*profile;

FILE	*edge_f;
FILE	*parse_f;
FILE	*run_f;
FILE	*result_f;

int	recording_edges = 0, recording_realizations = 0;
int	use_pinput, input_is_yy;

char	*ace_commandline = NULL;	// if we are running a single ace instead of connecting to an arbiter
int show_timing = 0;
int	appending = 0;

usage(char	*myname, int	status)
{
	fprintf(stderr, "usage: %s [ ARBITER_HOST | -a 'ace commandline' ] [ [-t] -e GOLD_PROFILE [ -p RID ] ] [ -f ] PROFILE\n", myname);
	fprintf(stderr, "  ARBITER_HOST  specifies an `arbiter' server where the processing will be done\n");
	fprintf(stderr, "  -a   runs a single parser/generator, instead of using `arbiter'\n");
	fprintf(stderr, "  -e   generation mode (default is parsing)\n");
	fprintf(stderr, "  -p   specify a specific result ID to generate from (default is to read `preference' relation)\n");
	fprintf(stderr, "  -t   transfer mode\n");
	fprintf(stderr, "  -f   record a forest-based profile (default is full derivations)\n");
	fprintf(stderr, "  -i   show show timing information\n");
	fprintf(stderr, "  -V   show version\n");
	fprintf(stderr, "  -h   show usage information\n");
	exit(status);
}

int verbose = 0;
int	transfer_mode = 0;
int	fabricate_preference = 0;
char	*fabricate_preferred_rid;
char	*only_iids;
struct hash	*only_iids_hash;

main(int	argc, char	*argv[])
{
	int	ch;
	setlocale(LC_ALL, "");
	setlocale(LC_NUMERIC, "POSIX");
	while( (ch = getopt(argc, argv, "tie:fa:VvhyYp:I:AF:")) != -1)switch(ch)
	{
		case	'V':
			printf("art version " VERSION "\n");
			exit(0);
		case	'h':
			usage(argv[0], 0);
		case	'v':
			verbose++;
			break;
		case	'f':
			recording_edges = 1;
			break;
		case	'i':
			show_timing = 1;
			break;
		case	'e':
			if(!transfer_mode)recording_realizations = 1;
			source_profile_path = optarg;
			break;
		case	'p':
			fabricate_preference = 1;
			fabricate_preferred_rid = optarg;
			break;
		case	't':
			transfer_mode = 1;
			recording_realizations = 0;
			break;
		case	'a':
			ace_commandline = optarg;
			break;
		case	'y':
			use_pinput = 1;
			input_is_yy = 1;
			break;
		case	'Y':
			input_is_yy = 1;
			break;
		case	'I': ;
			char	*iid;
			only_iids = optarg;
			if(!only_iids_hash)only_iids_hash = hash_new("iids");
			for(iid=strtok(only_iids," \n\t,");iid;iid=strtok(NULL, " \n\t,"))
				hash_add(only_iids_hash, strdup(iid), (void*)1L);
			break;
		case	'F': ;
			FILE	*iidf = fopen(optarg,"r");
			if(!iidf){perror(optarg);exit(-1);}
			if(!only_iids_hash)only_iids_hash = hash_new("iids");
			char	fiid[1024];
			while(fgets(fiid,1000,iidf))hash_add(only_iids_hash, strdup(trim_string(fiid)), (void*)1L);
			fclose(iidf);
			break;
		case	'A':
			appending = 1;
			break;
		default:
			usage(argv[0], -1);
	}
	if(ace_commandline)
	{
		if(argc-optind != 1)
			usage(argv[0], -1);
		int old_stderr = dup(2);
		pid_t p = forkpty(&arbiter_fd, NULL, NULL, NULL);
		if(p < 0)
		{
			fprintf(stderr, "unable to forkpty() to create an ACE process\n");
			exit(-1);
		}
		if(p==0)
		{
			char	command[10240];
			sprintf(command, "%s --tsdb-notes -m- --show-realization-trees --show-realization-mrses", ace_commandline);
			execl("/bin/bash", "bash", "-c", command, NULL);
			close(2);
			dup2(old_stderr, 2);
			perror("execl");
			exit(-1);
		}
		close(old_stderr);
		struct termios	tc;
		tcgetattr(arbiter_fd, &tc);
		cfmakeraw(&tc);
		tcsetattr(arbiter_fd, TCSANOW, &tc);
	}
	else
	{
		if(argc-optind != 2)
			usage(argv[0], -1);
		arbiter_host = argv[optind++];
		char	*colon = strchr(arbiter_host, ':');
		int	port = ARBITER_PORT;
		if(colon)
		{
			*colon = 0;
			port = atoi(colon+1);
		}
		arbiter_fd = tcpip_conn(arbiter_host, port);
		if(arbiter_fd <= 0)
		{
			fprintf(stderr, "unable to connect to arbiter daemon on %s:%d\n", arbiter_host, port);
			exit(-1);
		}
	}

	arbiter_f = fdopen(arbiter_fd, "r");
	arbiter_wrf = fdopen(arbiter_fd, "w");
	assert(arbiter_f != NULL);
	assert(arbiter_wrf != NULL);

	profile_path = argv[optind++];
	profile = load_tsdb(profile_path);
	if(!profile)
	{
		fprintf(stderr, "unable to load tsdb profile from %s\n", profile_path);
		exit(-1);
	}

	if(arbiter_host)	// use special arbiter command to display profile name in web ui
		fprintf(arbiter_wrf, "<<<&&&name %s\n", profile_path);

	struct relation	*item_r = get_relation(profile, "item");
	assert(item_r);
	int	item_i_id = get_field(item_r, "i-id", "integer");
	int item_i_input = get_field(item_r, "i-input", "string");
	struct relation	*parse = get_relation(profile, "parse");
	assert(parse);
	int parse_i_id = get_field(parse, "i-id", "integer");
	int parse_p_input = get_field(parse, "p-input", "string");
	int i;
	int nsubmitted = 0;

	if(source_profile_path)
	{
		source_profile = load_tsdb(source_profile_path);
		if(!source_profile)
		{
			fprintf(stderr, "unable to load tsdb source_profile from %s\n", source_profile_path);
			exit(-1);
		}
		// don't need compatibility, actually...
		//assert(tsdb_compatible(profile, source_profile));
		s_parse = get_relation(source_profile, "parse");
		s_preference = get_relation(source_profile, "preference");
		s_result = get_relation(source_profile, "result");
		assert(s_parse && s_preference && s_result);
	}

	char	**inputs;
	inputs = calloc(sizeof(char*),item_r->ntuples);
	for(i=0;i<item_r->ntuples;i++)
	{
		char	*i_id = item_r->tuples[i][item_i_id];
		if(only_iids_hash && !hash_find(only_iids_hash, i_id))continue;
		if(source_profile_path)
		{
			char	*mrs = get_source_mrs_for_i_id(i_id);
			if(!mrs)continue;
			inputs[i] = strdup(mrs);
		}
		else if(use_pinput)
		{
			char	**parse_tup = tsdb_lookup_relation_with_key(parse, parse_i_id, i_id);
			if(!parse_tup)continue;
			char	*pinput = parse_tup[parse_p_input];
			inputs[i] = strdup(pinput);
		}
		else
		{
			char	*i_input = item_r->tuples[i][item_i_input];
			inputs[i] = strdup(i_input);
		}
	}

int	submit_item(int	i)
{
	if(!inputs[i])return 0;
	if(verbose)printf("send: %s\n", inputs[i]);
	int res = fprintf(arbiter_wrf, "%s\n", inputs[i]);
	if(res < 0)
	{
		perror("writing to arbiter");
		exit(-1);
	}
	return 1;
}

	if(!ace_commandline)
	{
		// if using arbiter, send all the items ahead of time to allow out-of-order processing
		for(i=0;i<item_r->ntuples;i++)
			nsubmitted += submit_item(i);
		printf("%d / %d items submitted to arbiter\n", nsubmitted, item_r->ntuples);
		fflush(arbiter_wrf);
	}

	char	*run_id = "1";
	char	*platform_code = "art -- arbiter/tsdb recorder";
	for(i=0;i<item_r->ntuples;i++)
	{
		char	*i_id = item_r->tuples[i][item_i_id];
		char	*i_input = item_r->tuples[i][item_i_input];
		char	*parse_id = i_id;	// by fiat
		if(!inputs[i])
			continue;	// not expecting any result for this item
		if(ace_commandline)
			submit_item(i);	// submit items one-at-a-time if driving ACE directly
		fflush(arbiter_wrf);
		int status = read_result(parse_id, run_id, i_id, i_input);
		if(status != 0)
		{
			fprintf(stderr, "failed to read result for %s\n", i_id);
			exit(-1);
		}
	}
	record_run(run_id, platform_code, item_r->ntuples);
	if(run_f)fclose(run_f);
	if(parse_f)fclose(parse_f);
	if(result_f)fclose(result_f);
	if(edge_f)fclose(edge_f);
	if(show_timing)report_timers();
	return 0;
}

char	*get_source_mrs_for_i_id(char	*i_id)
{
	char	**ptup = tsdb_lookup_relation_with_key(s_parse, get_field(s_parse, "i-id", "integer"), i_id);
	if(!ptup) { fprintf(stderr, "no `parse' record for item %s\n", i_id); return NULL; }
	char	*p_id = ptup[get_field(s_parse, "parse-id", "integer")];
	char	*r_id;
	if(!fabricate_preference)
	{
		char	**preftup = tsdb_lookup_relation_with_key(s_preference, get_field(s_preference, "parse-id", "integer"), p_id);
		if(!preftup)return NULL;	// no prefered reading of this item; skip it.
		r_id = preftup[get_field(s_preference, "result-id", "integer")];
	}
	else r_id = fabricate_preferred_rid;
	char	*mrs = NULL;
	int	rid_field = get_field(s_result, "result-id", "integer");
	void	visitor(char	**rtup)
		{ if(!strcmp(rtup[rid_field], r_id))
			mrs = rtup[get_field(s_result, "mrs", "string")]; }
	tsdb_visit_relation_with_key(s_result, get_field(s_result, "parse-id", "integer"), p_id, visitor);
	if(!mrs) { fprintf(stderr, "no `result' record for item %s parse %s result %s\n", i_id, p_id, r_id); return NULL; }
	if(!*mrs) { fprintf(stderr, "`mrs' field of `result' relation is empty for item %s parse %s result %s\n", i_id, p_id, r_id); return NULL; }
	return mrs;
}

char	*parse_info_string;
char	*run_info_string;

int	read_result(char	*parse_id, char	*run_id, char	*i_id, char	*i_input)
{
	const int		buffer_len = 1024*1024;
	char	*buffer = malloc(buffer_len);
	int	got_header = 0;
	int	nresults = 0;
	int	nedges[5] = {0,0,0,0,0};
	int	edge_is_root = -1;
	int	root_edge_id_gen = -1;
	int	got_prob = 0;
	double	prob;
	char	*last_gen_dtree = NULL, *last_gen_mrs = NULL, *last_gen_surface = NULL;
	printf("reading results for %16s	", i_id); fflush(stdout);
	if(parse_info_string)free(parse_info_string);
	parse_info_string = NULL;

	if(transfer_mode)got_header = 1;

	static int read_all_result_timer = -1, read_timer = -1, record_result_timer = -1;
	static int strlen_timer = -1, strstr_timer = -1;
	if(read_timer == -1)
	{
		read_all_result_timer = new_timer("read_result (all)");
		read_timer = new_timer("fgets results");
		record_result_timer = new_timer("record_result");
		strlen_timer = new_timer("read_result: strlen");
		strstr_timer = new_timer("read_result: strstr");
	}

	start_timer(read_all_result_timer);
	while(1)
	{
		start_timer(read_timer);
		char	*rv = fgets(buffer, buffer_len-1, arbiter_f);
		stop_timer(read_timer, 1);
		if(rv == NULL)break;
		if(!strncmp(buffer, "WARNING: ", 9))
			continue;
		if(!strncmp(buffer, "NOTE: ", 6))
		{
			//printf("ignoring %s", buffer);
			if(!strncmp(buffer, "NOTE: tsdb parse:", 17))
				parse_info_string = strdup(buffer);
			if(!strncmp(buffer, "NOTE: tsdb run:", 15))
				run_info_string = strdup(buffer);
			continue;
		}

		start_timer(strlen_timer);
		size_t	len = strlen(buffer);
		stop_timer(strlen_timer, 1);
		if(len == 0)
		{
			perror("unexpected end of stream from arbiter");
			return -1;
		}
		char	*end = buffer+len-1;
		*end = 0;
		if(verbose)printf("read: `%s'\n", buffer);
		if(!strncmp(buffer, "SENT: ", 6) || !strncmp(buffer, "SKIP: ", 6))
		{
			if(strcmp(buffer+6, i_input) && !recording_realizations && !transfer_mode && !input_is_yy)
			{
				fprintf(stderr, "out of sync; arbiter sent wrong sentence.\n");
				fprintf(stderr, "expected: %s\n", i_input);
				fprintf(stderr, "received: %s\n", buffer+6);
				return -1;
			}
			if(input_is_yy)assert(!strcmp(buffer+6, "(yy mode)"));
			got_header = 1;
			continue;
		}
		if(*buffer == 0)
		{
			if(!got_header)continue; // stray newline
			else break;	// end of item marker
		}
		if(!got_header)
		{
			fprintf(stderr, "out of sync; arbiter sent '%s' when expecting a SENT or SKIP header.\n", buffer);
			return -1;
		}
		if(!strncasecmp(buffer, "probability = ", 14))
		{
			got_prob = 1;
			prob = atof(buffer+14);
			continue;
		}
		if(recording_edges)
		{
			// recording packed forest and token lattice
			// each line returned from arbiter will be one of:
			// root <pseudo-edge-id> <from> <to> 123 root_strict
			// token -8 0 1 <0-or-1-usedness> token "AVM" string...
			// (123 sign 0 13 was_packed <lexeme=1,lexrule=2,synrule=3> [comma,separated,packed,edge,ids] (comma,separated,daughter,ids)
			int	id, from, to, offset, used, did, status, type;
			char	roottype[1024];
			if(5 == sscanf(buffer, "root %d %d %d %d %s", &id, &from, &to, &did, roottype))
			{
				// root edge specifier
				edge_is_root = did;
				char	dlist[256];
				sprintf(dlist, "(%d)", did);
				record_edge(parse_id, id, roottype, from, to, ROOT_EDGE_TYPE, RESULT_EDGE_STATUS, dlist, "()");
				nedges[ROOT_EDGE_TYPE]++;
			}
			else if(3 <= sscanf(buffer, "token %d %d %d %d %n", &id, &from, &to, &used, &offset))
			{
				// token description
				char	*avm = buffer + offset;
				record_edge(parse_id, id, avm, from, to, TOKEN_EDGE_TYPE, used?RESULT_EDGE_STATUS:0, "()", "()");
				nedges[TOKEN_EDGE_TYPE]++;
			}
			else
			{
				// edge description
				char	*plist = strchr(buffer+1, '[');
				char	*dlist = strchr(buffer+1, '(');
				if(!plist || !dlist)
				{
					fprintf(stderr, "malformed edge `%s'\n", buffer);
					return -1;
				}
				char	sign[1024];
				if(6 != sscanf(buffer, "(%d %1020[^ ] %d %d %d %d [", &id, sign, &from, &to, &status, &type))
				{
					fprintf(stderr, "malformed edge `%s'\n", buffer);
					return -1;
				}
				// convert packing and daughter comma-separated lists to lisp-style space-separated lists
				*plist = *dlist = '(';
				char	*c;
				for(c=plist+1;*c && *c!=']';c++)
					if(*c==',')*c=' ';
				if(!*c) { fprintf(stderr, "undelimited packing list\n"); return -1; }
				*c++ = ')'; *c++ = 0;
				for(c=dlist+1;*c && *c!=')';c++)
					if(*c==',')*c=' ';
				if(!*c) { fprintf(stderr, "undelimited daughter list\n"); return -1; }
				*c++ = ')'; *c++ = 0;
				record_edge(parse_id, id, sign, from, to, type, status, dlist, plist);
				assert(type>=0 && type<=4);
				nedges[type]++;
			}
		}
		else if(recording_realizations)
		{
			// recording discrete generation results
			// each line returned from arbiter will look one of:
			// DTREE = ( derivation )
			// MRS = [ mrs ]
			// SURFACE = The dog chased the cat.
			// END RESULT
			if(!strncmp(buffer, "DTREE = ", 8))
				last_gen_dtree = strdup(buffer + 8);
			else if(!strncmp(buffer, "MRS = ", 6))
				last_gen_mrs = strdup(buffer + 6);
			else if(!strncmp(buffer, "SURFACE = ", 10))
				last_gen_surface = strdup(buffer + 10);
			else if(!strcmp(buffer, "END RESULT"))
			{
				record_result(parse_id, nresults++, last_gen_dtree?:"", last_gen_mrs?:"", last_gen_surface?:"", got_prob, prob);
				if(last_gen_dtree)free(last_gen_dtree); last_gen_dtree = NULL;
				if(last_gen_mrs)free(last_gen_mrs); last_gen_mrs = NULL;
				if(last_gen_surface)free(last_gen_surface); last_gen_surface = NULL;
			}
			else
			{
				fprintf(stderr, "malformed generation result; neither dtree nor mrs nor surface\n");
				return -1;
			}
			got_prob = 0;
		}
		else if(transfer_mode)
		{
			// recording discrete mrs results
			// each line returned from arbiter will look like:
			// [ mrs ]
			start_timer(record_result_timer);
			record_result(parse_id, nresults, "", buffer, "", got_prob, prob);
			stop_timer(record_result_timer, 1);
			nresults++;
			got_prob = 0;
		}
		else
		{
			// recording discrete parse results
			// each line returned from arbiter will look like:
			// [ mrs ] ; ( derivation )
			start_timer(strstr_timer);
			char	*sep = strstr(buffer, "] ;  (");
			stop_timer(strstr_timer, 1);
			if(!sep)
			{
				fprintf(stderr, "malformed result; no `] ;  (' separator\n");
				fprintf(stderr, "offending result:\n%s\n", buffer);
				return -1;
			}
			sep[2] = 0;
			char	*mrs_string = buffer;
			char	*dtree_string = sep+5;
			//printf("MRS string = `%s'\n", mrs_string);
			//printf("Dtree string = `%s'\n", dtree_string);
			start_timer(record_result_timer);
			record_result(parse_id, nresults, dtree_string, mrs_string, "", got_prob, prob);
			stop_timer(record_result_timer, 1);
			nresults++;
			got_prob = 0;
		}
	}
	record_parse(parse_id, run_id, i_id, nresults, nedges);
	printf("%d results\n", nresults);
	stop_timer(read_all_result_timer, nresults);
	return 0;
}

write_tuple(FILE	*f, char	**tuple, struct relation	*r)
{
	int i;

	// supply default values
	for(i=0;i<r->nfields;i++)
	{
		if(!*tuple[i])
		{
			char	*type = r->fields[i].type;
			if(!strcasecmp(type, "integer"))tuple[i] = strdup("-1");
			else if(!strcasecmp(type, "string"))tuple[i] = strdup("");
			else if(!strcasecmp(type, "date"))tuple[i] = strdup("23-6-2013 14:28:24");	// XXX fix me
		}
	}

	for(i=0;i<r->nfields;i++)
	{
		if(i)fprintf(f, "@");
		char	*encode_tsdb_field(char	*plain);
		if(*tuple[i])
		{
			char	*esc = encode_tsdb_field(tuple[i]);
			fprintf(f, "%s", esc);
			free(esc);
		}
	}
	fprintf(f, "\n");
}

void	lispy_get_field(char	*buffer, char	*field, char	*into)
{
	*into=0;
	if(!buffer)return;
	buffer = strstr(buffer, field);
	if(!buffer)return;
	buffer += strlen(field);
	buffer = strchr(buffer, '.');
	if(!buffer)return;
	buffer++;
	char	*rp = strchr(buffer, ')');
	if(!buffer)return;
	rp--;
	while(rp>buffer && *rp==' ' || *rp=='\t')rp--;
	rp++;
	while(*buffer==' ' || *buffer=='\t')buffer++;
	if(*buffer=='"' && rp[-1]=='"')
		{ buffer++; rp--; }
	memcpy(into, buffer, rp-buffer);
	into[rp-buffer]=0;
}

FILE	*open_output_file(char	*dir, char	*name)
{
	FILE	*f = NULL;
	char	path[1024];
	sprintf(path, "%s/%s.gz", dir, name);
	unlink(path);
	sprintf(path, "%s/%s", dir, name);
	if(appending)
	{
		f = fopen(path, "a");
	}
	else
	{
		unlink(path);
		f = fopen(path, "w");
	}
	if(!f)
	{
		perror(path);
		exit(-1);
	}
	return f;
}

void	record_parse(char	*parse_id, char	*run_id, char	*i_id, int	nresults, int	*nedges)
{
	static int init = 0;
	static struct relation	*parse_r;
	static int	parse_parse_id, parse_run_id, parse_i_id, parse_readings;
	static char	**tuple;
	if(!init)
	{
		parse_r = get_relation(profile, "parse");
		assert(parse_r);
		parse_parse_id = get_field(parse_r, "parse-id", "integer");
		parse_run_id = get_field(parse_r, "run-id", "integer");
		parse_i_id = get_field(parse_r, "i-id", "integer");
		parse_readings = get_field(parse_r, "readings", "integer");
		tuple = malloc(sizeof(char*)*parse_r->nfields);
		int i;
		for(i=0;i<parse_r->nfields;i++)tuple[i] = "";
		parse_f = open_output_file(profile->path, "parse");
		init = 1;
	}

	tuple[parse_parse_id] = parse_id;
	tuple[parse_run_id] = run_id;
	tuple[parse_i_id] = i_id;
	char	readings[128];
	sprintf(readings, "%d", nresults);
	tuple[parse_readings] = readings;

	char	total[128], tcpu[128], treal[128], others[128];
	tuple[get_field(parse_r, "total", "integer")] = total;
	lispy_get_field(parse_info_string, ":total", total);
	tuple[get_field(parse_r, "tcpu", "integer")] = tcpu;
	lispy_get_field(parse_info_string, ":tcpu", tcpu);
	tuple[get_field(parse_r, "treal", "integer")] = treal;
	lispy_get_field(parse_info_string, ":treal", treal);
	tuple[get_field(parse_r, "others", "integer")] = others;
	lispy_get_field(parse_info_string, ":others", others);

	char	tedges[128], eedges[128], ledges[128], sedges[128], redges[128];
	sprintf(tuple[get_field(parse_r, "tedges", "integer")]=tedges, "%d", nedges[0]);
	sprintf(tuple[get_field(parse_r, "eedges", "integer")]=eedges, "%d", nedges[1]);
	sprintf(tuple[get_field(parse_r, "ledges", "integer")]=ledges, "%d", nedges[2]);
	sprintf(tuple[get_field(parse_r, "sedges", "integer")]=sedges, "%d", nedges[3]);
	sprintf(tuple[get_field(parse_r, "redges", "integer")]=redges, "%d", nedges[4]);

	char	error[1024] = "";
	tuple[get_field(parse_r, "error", "string")] = error;
	lispy_get_field(parse_info_string, ":error", error);

	static int write_timer = -1;
	if(write_timer == -1)write_timer = new_timer("parse table write");
	start_timer(write_timer);
	write_tuple(parse_f, tuple, parse_r);
	stop_timer(write_timer, 1);
}

void	record_run(char	*run_id, char	*platform, int	nitems)
{
	static int init = 0;
	static struct relation	*run_r;
	static int	run_run_id, run_platform, run_items;
	static char	**tuple;
	if(!init)
	{
		run_r = get_relation(profile, "run");
		assert(run_r);
		run_run_id = get_field(run_r, "run-id", "integer");
		run_platform = get_field(run_r, "platform", "string");
		run_items = get_field(run_r, "items", "integer");
		tuple = malloc(sizeof(char*)*run_r->nfields);
		int i;
		for(i=0;i<run_r->nfields;i++)tuple[i] = "";
		run_f = open_output_file(profile->path, "run");
		init = 1;
	}

	tuple[run_run_id] = run_id;
	tuple[run_platform] = platform;
	char	items[128];
	sprintf(items, "%d", nitems);
	tuple[run_items] = items;

	// lispy fields: (:application . "answer") (:platform . "gcc 4.4") (:grammar . "ERG (trunk)") (:avms . 8566) (:lexicon . 37919) (:lrules . 77) (:rules . 219)
	char	*fields[] = {":application", ":grammar", ":avms", ":lexicon", ":lrules", ":rules", NULL};
	char	*types[] = {"string", "string", "integer", "integer", "integer", "integer"};
	char	buffers[10][128];
	int j;
	for(j=0;fields[j];j++)
	{
		tuple[get_field(run_r, fields[j]+1, types[j])] = buffers[j];
		lispy_get_field(run_info_string, fields[j], buffers[j]);
	}

	write_tuple(run_f, tuple, run_r);
}

void	record_result(char	*parse_id, int	result_idx, char	*derivation, char	*mrs, char	*surface, int	got_prob, double	prob)
{
	static int init = 0;
	static struct relation	*result_r;
	static int	result_parse_id, result_result_id, result_derivation, result_mrs, result_surface, result_flags;
	static char	**tuple;
	if(!init)
	{
		result_r = get_relation(profile, "result");
		assert(result_r);
		result_parse_id = get_field(result_r, "parse-id", "integer");
		result_result_id = get_field(result_r, "result-id", "integer");
		result_derivation = get_field(result_r, "derivation", "string");
		result_mrs = get_field(result_r, "mrs", "string");
		result_surface = get_field(result_r, "surface", "string");
		result_flags = get_field(result_r, "flags", "string");
		tuple = malloc(sizeof(char*)*result_r->nfields);
		int i;
		for(i=0;i<result_r->nfields;i++)tuple[i] = "";
		result_f = open_output_file(profile->path, "result");
		init = 1;
	}

	tuple[result_parse_id] = parse_id;
	char	result_id[128];
	sprintf(result_id, "%d", result_idx);
	tuple[result_result_id] = result_id;
	tuple[result_derivation] = derivation;
	tuple[result_mrs] = mrs;
	tuple[result_surface] = surface;
	char	prob_str[128] = "";
	if(got_prob)sprintf(prob_str, "%f", prob);
	tuple[result_flags] = prob_str;
	write_tuple(result_f, tuple, result_r);
}

void	record_edge(char	*parse_id, int	id, char	*sign, int	from, int	to, int	type, int	status, char	*dlist, char	*plist)
{
	static int init = 0;
	static struct relation	*edge_r;
	static int	edge_parse_id, edge_e_id, edge_e_label, edge_e_start, edge_e_end, edge_e_status, edge_e_type, edge_e_daughters, edge_e_alternates;
	static char	**tuple;
	if(!init)
	{
		edge_r = get_relation(profile, "edge");
		assert(edge_r);
		edge_parse_id = get_field(edge_r, "parse-id", "integer");
		edge_e_id = get_field(edge_r, "e-id", "integer");
		edge_e_start = get_field(edge_r, "e-start", "integer");
		edge_e_end = get_field(edge_r, "e-end", "integer");
		edge_e_status = get_field(edge_r, "e-status", "integer");
		edge_e_type = get_field(edge_r, "e-type", "integer");
		edge_e_label = get_field(edge_r, "e-label", "string");
		edge_e_daughters = get_field(edge_r, "e-daughters", "string");
		edge_e_alternates = get_field(edge_r, "e-alternates", "string");
		tuple = malloc(sizeof(char*)*edge_r->nfields);
		int i;
		for(i=0;i<edge_r->nfields;i++)tuple[i] = "";
		edge_f = open_output_file(profile->path, "edge");
		init = 1;
	}

	tuple[edge_parse_id] = parse_id;

	char	edge_id[128];
	sprintf(edge_id, "%d", id);
	tuple[edge_e_id] = edge_id;

	char	edge_from[128];
	sprintf(edge_from, "%d", from);
	tuple[edge_e_start] = edge_from;

	char	edge_to[128];
	sprintf(edge_to, "%d", to);
	tuple[edge_e_end] = edge_to;

	char	edge_status[128];
	sprintf(edge_status, "%d", status);
	tuple[edge_e_status] = edge_status;

	char	edge_type[128];
	sprintf(edge_type, "%d", type);
	tuple[edge_e_type] = edge_type;

	tuple[edge_e_label] = sign;
	tuple[edge_e_daughters] = (!strcmp(dlist, "()"))?"":dlist;
	tuple[edge_e_alternates] = (!strcmp(plist, "()"))?"":plist;

	write_tuple(edge_f, tuple, edge_r);
}
