#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>
#include	<string.h>

#include	<getopt.h>

#include	<tsdb.h>

#include	<ace/tree.h>
#include	<ace/reconstruct.h>

char	*get_source_mrs_for_i_id(char	*i_id);

char	*profile_path;
struct tsdb	*profile;

int	reconstruct_all = 0, check_mrs = 0, show_avm = 0;
char	*grammar_path = NULL;
char	*detail_iid = NULL;

int	n_total, n_dparsed, n_drecons, n_mparsed, n_mextract, n_mmismatch;

mrs_detail(struct mrs	*smrs, struct mrs	*emrs)
{
	extern char mrs_tab;
	mrs_tab = '\n';
	printf("stored MRS:\n");
	print_mrs(smrs);
	printf("reconstructed MRS:\n");
	print_mrs(emrs);
}

char	*reconstruct_result(char	*i_id, char	*i_input, char	*derivation, char	*stored_mrs)
{
	n_total++;

	clear_slab();
	clear_mrs();

	derivation = strdup(derivation);
	struct tree	*t = string_to_tree(derivation);
	free(derivation);
	if(!t)return "derivation format error";

	n_dparsed++;

	struct dg	*avm = reconstruct_tree(t, NULL);
	if(!avm)return "reunification failed";

	n_drecons++;

	if(show_avm)
	{
		print_dg(avm);
		printf("\n");
	}

	if(detail_iid)
		printf("reconstructed successfully.\n");
	if(!check_mrs)return NULL;

	struct mrs	*smrs = parse_mrs(stored_mrs);
	if(smrs)n_mparsed++;
	struct mrs	*emrs = extract_mrs(avm);
	if(emrs)n_mextract++;
	if(!smrs)return "stored MRS unreadable";
	if(!emrs)return "internal MRS error";

	if(!mrs_subsumes(smrs, emrs))
	{
		n_mmismatch++;
		if(detail_iid)
			mrs_detail(smrs, emrs);
		return "stored MRS is more specific";
	}
	if(!mrs_subsumes(emrs, smrs))
	{
		n_mmismatch++;
		if(detail_iid)
			mrs_detail(smrs, emrs);
		return "reconstructed MRS is more specific";
	}
	char	*r = is_root(avm);
	if(detail_iid)
		printf("no differences; reconstructed root symbol = %s.\n", r);
	if(!r)return "no matching root symbol";
	return NULL;
}

main(int	argc, char	*argv[])
{
	int	ch;
	int verbosity = 0;
	int one_tree = 0;
	while( (ch = getopt(argc, argv, "g:ami:vtf")) != -1)switch(ch)
	{
		case	'a': reconstruct_all = 1; break;
		case	'm': check_mrs = 1; break;
		case	'g': grammar_path = optarg; break;
		case	'i': detail_iid = optarg; break;
		case	'v': verbosity++; break;
		case	't':	one_tree = 1; break;
		case	'f': show_avm++;	break;
		default:
			fprintf(stderr, "unknown option\n");
			exit(-1);
	}
	if(!((one_tree && argc-optind==0) || (!one_tree && argc-optind == 1)) || !grammar_path)
	{
		fprintf(stderr, "usage: %s -g grammar.dat [-v] [-f] { -t | [-a] [-m] [-i detail-i-id] my-profile }\n", argv[0]);
		exit(-1);
	}

	ace_load_grammar(grammar_path);
	if(show_avm > 1)
	{
		extern void	*deleted_daughters, *copy_restrictor, *unify_restrictor;
		deleted_daughters = copy_restrictor = unify_restrictor = NULL;
	}

	reconstruction_noise_level = verbosity;

	if(one_tree)
	{
		char	buffer[102400];
		detail_iid = "1";
		assert(NULL != fgets(buffer, 102399, stdin));
		char	*message = reconstruct_result("1", "console", buffer, "");
		if(message)printf("message: %s\n", message);
		return 0;
	}

	profile_path = argv[optind++];
	profile = load_tsdb(profile_path);
	if(!profile)
	{
		fprintf(stderr, "unable to load tsdb profile from %s\n", profile_path);
		exit(-1);
	}

	struct relation	*item_r = get_relation(profile, "item");
	assert(item_r);
	int	item_i_id = get_field(item_r, "i-id", "integer");
	int item_i_input = get_field(item_r, "i-input", "string");
	int i;

	struct relation	*parse = get_relation(profile, "parse");
	struct relation	*preference = get_relation(profile, "preference");
	struct relation	*result = get_relation(profile, "result");
	assert(parse && preference && result);

	int	parse_i_id = get_field(parse, "i-id", "integer");
	int	parse_parse_id = get_field(parse, "parse-id", "integer");
	int	preference_parse_id = get_field(preference, "parse-id", "integer");
	int	preference_result_id = get_field(preference, "result-id", "integer");
	int	result_parse_id = get_field(result, "parse-id", "integer");
	int	result_result_id = get_field(result, "result-id", "integer");
	int	result_derivation = get_field(result, "derivation", "string");
	int	result_mrs = get_field(result, "mrs", "string");

	for(i=0;i<item_r->ntuples;i++)
	{
		char	*i_id = item_r->tuples[i][item_i_id];
		if(detail_iid && strcmp(detail_iid, i_id))continue;

		char	*i_input = item_r->tuples[i][item_i_input];
		if(detail_iid)
		{
			printf("\n");
			printf("details on item %s\n", i_id);
			printf("%s\n", i_input);
			printf("\n");
		}

		char	**ptup = tsdb_lookup_relation_with_key(parse, parse_i_id, i_id);
		if(!ptup) { fprintf(stderr, "no `parse' record for item %s\n", i_id); exit(-1); }
		char	*p_id = ptup[parse_parse_id];
		char	*only_r_id = NULL;

		if(!reconstruct_all)
		{
			char	**preftup = tsdb_lookup_relation_with_key(preference, preference_parse_id, p_id);
			if(!preftup)continue;	// no prefered reading of this item; skip it.
			only_r_id = preftup[preference_result_id];
		}
		void visitor(char	**rtup)
		{
			if(!only_r_id || !strcmp(only_r_id, rtup[result_result_id]))
			{
				char	*message = reconstruct_result(i_id, i_input, rtup[result_derivation], rtup[result_mrs]);
				if(message)
				{
					printf("item %s result %s	%s\n", i_id, rtup[result_result_id], message);
					if(verbosity && !detail_iid)printf("\n");
				}
			}
		}
		tsdb_visit_relation_with_key(result, result_parse_id, p_id, visitor);

		if(detail_iid)
			printf("\n");
	}

	printf("%-30s: %d found\n",
		"`result' tuples", n_total);
	printf("%-30s: %d format errors\n",
		"derivations", n_total - n_dparsed);
	printf("%-30s: %d conflicts\n",
		"reunified", n_dparsed - n_drecons);
	if(check_mrs)
	{
		printf("%-30s: %d internal errors\n",
			"extracted MRSes", n_drecons - n_mextract);
		printf("%-30s: %d format errors\n",
			"stored MRSes", n_total - n_mparsed);
		printf("%-30s: %d mismatches\n",
			"compared MRSes", n_mmismatch);
	}
}
