#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<assert.h>
#include	<errno.h>

#include	<ace/hash.h>

#include	"tsdb.h"

char	delimchar = '@';

struct tsdb	*load_tsdb(char	*path)
{
	struct tsdb	*t = calloc(sizeof(*t), 1);
	t->path = strdup(path);
	read_relations(t, path);
	return t;
}

void	tsdb_free_profile(struct tsdb	*profile)
{
	free(profile->path);
	int i, j, k;
	for(i=0;i<profile->nrelations;i++)
	{
		struct relation	*r = profile->relations + i;
		free(r->name);
		if(r->loaded)
		{
			for(j=0;j<r->ntuples;j++)
			{
				// all the values in tuples[j][] are in the same allocated buffer
				free(r->tuples[j][0]);
				//for(k=0;k<r->nfields;k++)
					//if(!r->fields[k].iskey)	// was here because hash_free() took care of these; commented out because hash_free_nokeys() doesn't take care of these.
					//free(r->tuples[j][k]);
				free(r->tuples[j]);
			}
			free(r->tuples);
		}
		for(j=0;j<r->nfields;j++)
		{
			free(r->fields[j].name);
			free(r->fields[j].type);
			if(r->fields[j].iskey && r->loaded)
				hash_free_nokeys(r->fields[j].hash);
		}
		free(r->fields);
	}
	free(profile->relations);
	free(profile);
}

static void	load_tsdb_table(char	*home, char	*fname, int	(*loader)(char	*line))
{
	char	*full_fname = malloc(strlen(home) + strlen(fname) + 10);
	sprintf(full_fname, "%s/%s", home, fname);

	FILE	*f;
	f = fopen(full_fname, "r");
	if(!f)
	{
		strcat(full_fname, ".gz");
		f = fopen(full_fname, "r");
		if(!f)
		{
			perror(full_fname);
			exit(-1);
		}
		fclose(f);
		char	cmd[1024];
		sprintf(cmd, "zcat %s", full_fname);
		f = popen(cmd, "r");
	}

	int	line_num = 0;
	int		blen = 1024*1024;
	char	*buffer = malloc(1024*1024);
	while(!feof(f))
	{
		line_num++;
		if(NULL == fgets(buffer, blen-10, f))
			break;
		char	*nl = buffer + strlen(buffer)-1;
		if(nl >= buffer && *nl=='\n')*nl-- = 0;
		if(nl >= buffer && *nl=='\r')*nl-- = 0;
		if(0 != loader(buffer))
			{ fprintf(stderr, "unexpected error in tsdb file %s:%d\n", full_fname, line_num); exit(-1); }
	}
	free(buffer);

	if(strstr(full_fname, ".gz"))
		pclose(f);
	else fclose(f);
	free(full_fname);
}

char	*encode_tsdb_field(char	*raw)
{
	char	*v = malloc(strlen(raw)*2 + 1), *result = v;
	while(*raw)
	{
		char	ec = 0;
		if(*raw == delimchar)ec = 's';
		else if(*raw == '\n')ec = 'n';
		else if(*raw == '\t')ec = 't';
		else if(*raw == '\r')ec = 'r';
		else if(*raw == '\\')ec = '\\';
		if(ec)
		{
			*v++ = '\\';
			*v++ = ec;
		}
		else *v++ = *raw;
		raw++;
	}
	*v = 0;
	return result;
}

char	*decode_tsdb_field(char	*code)
{
	//char	*v = malloc(strlen(code) + 1), *result = v;
	char	*v = code, *result = v;
	while(*code)
	{
		if(*code == '\\')
		{
			if(code[1] == 0)
			{
				fprintf(stderr, "bad escape sequence\n");
				exit(-1);
			}
			switch(code[1])
			{
				case	's':	*v++ = delimchar;	break;
				case	'n':	*v++ = '\n';	break;
				case	't':	*v++ = '\t';	break;
				case	'r':	*v++ = '\r';	break;
				case	'\\':	*v++ = '\\';	break;
				default:
					fprintf(stderr, "bad escape sequence\n");
					exit(-1);
					break;
			}
			code += 2;
		}
		else *v++ = *code++;
	}
	*v = 0;
	return result;
}

int	load_one_tsdb_record(struct relation	*r, char	*line)
{
	int	f = 0;
	char	*value, **tuple = calloc(sizeof(char*), r->nfields);
	line = strdup(line);
	for(f = 0; f < r->nfields; f++)
	{
		if(!line) return -1;
		value = line;
		line = strchr(line, delimchar);
		if(line)*line++ = 0;
		tuple[f] = decode_tsdb_field(value);
	}
	if(line)return -1;
	if(r->ntuples >= r->atuples)
	{
		r->atuples = (r->ntuples + 1) * 2;
		r->tuples = realloc(r->tuples, sizeof(char**) * r->atuples);
	}
	r->ntuples++;
	r->tuples[r->ntuples-1] = tuple;
	return 0;
}

void	index_relation(struct relation	*r)
{
	int	i, j;
	for(i=0;i<r->nfields;i++)
	{
		struct field	*f = &r->fields[i];
		if(!f->iskey)continue;
		f->hash = hash_new(f->name);
		for(j=0;j<r->ntuples;j++)
			hash_add(f->hash, r->tuples[j][i], r->tuples[j]);
	}
}

void	load_relation_data(struct tsdb	*p, struct relation	*r)
{
	int	loader(char	*line) { return load_one_tsdb_record(r, line); }
	load_tsdb_table(p->path, r->name, loader);
	index_relation(r);
	r->loaded = 1;
}

void	tsdb_reindex_relation(struct relation	*r)
{
	int i;
	for(i=0;i<r->nfields;i++)
	{
		struct field	*f = &r->fields[i];
		if(!f->iskey)continue;
		if(f->hash)hash_free_nokeys(f->hash);
	}
	index_relation(r);
}

int	tsdb_write_relation(struct tsdb	*p, struct relation	*r)
{
	if(!r->loaded) { fprintf(stderr, "cannot write out '%s' which is not loaded\n", r->name); return -1; }

	char	fname[10240];
	char	fname_bk[10240];
	sprintf(fname, "%s/%s", p->path, r->name);
	sprintf(fname_bk, "%s/.%s.save", p->path, r->name);
	int res = rename(fname, fname_bk);
	if(res != 0 && errno != ENOENT)
	{
		perror("rename");
		return -1;
	}
	FILE	*f = fopen(fname, "w");
	if(!f)
	{
		perror(fname);
		res = rename(fname_bk, fname);
		if(res != 0) { perror("rename"); fprintf(stderr, "leaving relation backup at '%s'\n", fname_bk); }
		return -1;
	}

	printf("writing tsdb relation '%s' with %d tuples\n", r->name, r->ntuples);

	int i, j;
	for(i=0;i<r->ntuples;i++)
	{
		for(j=0;j<r->nfields;j++)
		{
			char	*code = encode_tsdb_field(r->tuples[i][j]);
			if(j)
			{
				res = fprintf(f, "%c", delimchar);
				if(res < 0) { perror("fprintf"); goto fail; }
			}
			res = fprintf(f, "%s", code);
			if(res < 0) { perror("fprintf"); goto fail; }
			free(code);
		}
		res = fprintf(f, "\n");
		if(res < 0) { perror("fprintf"); goto fail; }
	}
	res = fflush(f);
	if(res != 0) { perror("fflush"); goto fail; }
	fclose(f);
	unlink(fname_bk);
	return 0;
fail:
	fclose(f);
	unlink(fname);
	res = rename(fname_bk, fname);
	if(res != 0) { perror("rename"); fprintf(stderr, "leaving relation backup at '%s'\n", fname_bk); }
	return -1;
}

int	tsdb_visit_relation_with_key(struct relation	*r, int	keyfield, char	*keyvalue, void	(*visitor)(char	**tuple))
{
	assert(r->fields[keyfield].iskey);
	hash_visit_key(r->fields[keyfield].hash, keyvalue, (void(*)(void*))visitor);
}

char	**tsdb_lookup_relation_with_key(struct relation	*r, int	keyfield, char	*keyvalue)
{
	assert(r->fields[keyfield].iskey);
	return hash_find(r->fields[keyfield].hash, keyvalue);
}
